// src/ts.ts
import { readFile } from "fs/promises";
import { existsSync } from "fs";
import path from "path";
import { babelParse, getLang, resolveObjectKey } from "@vue-macros/common";
import { isDeclaration } from "@babel/types";
var tsFileCache = {};
async function getTSFile(filePath) {
  if (tsFileCache[filePath])
    return tsFileCache[filePath];
  const content = await readFile(filePath, "utf-8");
  const program = babelParse(
    await readFile(filePath, "utf-8"),
    getLang(filePath)
  );
  return tsFileCache[filePath] = {
    filePath,
    content,
    ast: program.body
  };
}
var isTSDeclaration = (node) => isDeclaration(node) && node.type.startsWith("TS");
function mergeTSProperties(a, b) {
  return {
    callSignatures: [...a.callSignatures, ...b.callSignatures],
    constructSignatures: [...a.constructSignatures, ...b.constructSignatures],
    methods: { ...a.methods, ...b.methods },
    properties: { ...a.properties, ...b.properties }
  };
}
async function resolveTSProperties({
  type,
  file
}) {
  if (type.type === "TSInterfaceBody") {
    return resolveTypeElements(file, type.body);
  } else if (type.type === "TSTypeLiteral") {
    return resolveTypeElements(file, type.members);
  } else if (type.type === "TSInterfaceDeclaration") {
    let properties = resolveTypeElements(file, type.body.body);
    if (type.extends) {
      const resolvedExtends = (await Promise.all(
        type.extends.map(
          (node) => node.expression.type === "Identifier" ? resolveTSReferencedType({
            file,
            type: node.expression
          }) : void 0
        )
      )).filter(filterValidExtends);
      if (resolvedExtends.length > 0) {
        const ext = (await Promise.all(
          resolvedExtends.map((resolved) => resolveTSProperties(resolved))
        )).reduceRight((acc, curr) => mergeTSProperties(acc, curr));
        properties = mergeTSProperties(ext, properties);
      }
    }
    return properties;
  } else if (type.type === "TSIntersectionType") {
    let properties = {
      callSignatures: [],
      constructSignatures: [],
      methods: {},
      properties: {}
    };
    for (const subType of type.types) {
      const resolved = await resolveTSReferencedType({
        file,
        type: subType
      });
      if (!filterValidExtends(resolved))
        continue;
      properties = mergeTSProperties(
        properties,
        await resolveTSProperties(resolved)
      );
    }
    return properties;
  } else {
    throw new Error(`unknown node: ${type == null ? void 0 : type.type}`);
  }
  function filterValidExtends(node) {
    return [
      "TSInterfaceDeclaration",
      "TSTypeLiteral",
      "TSIntersectionType"
    ].includes(node == null ? void 0 : node.type.type);
  }
}
function resolveTypeElements(file, elements) {
  var _a;
  const properties = {
    callSignatures: [],
    constructSignatures: [],
    methods: {},
    properties: {}
  };
  const tryGetKey = (element) => {
    try {
      return resolveObjectKey(element.key, element.computed, false);
    } catch {
    }
  };
  for (const element of elements) {
    switch (element.type) {
      case "TSCallSignatureDeclaration":
        properties.callSignatures.push({ file, type: element });
        break;
      case "TSConstructSignatureDeclaration":
        properties.constructSignatures.push({ file, type: element });
        break;
      case "TSMethodSignature": {
        const key = tryGetKey(element);
        if (!key)
          continue;
        if (properties.properties[key])
          continue;
        if (!properties.methods[key])
          properties.methods[key] = [];
        if (element.typeAnnotation) {
          properties.methods[key].push({ file, type: element });
        }
        break;
      }
      case "TSPropertySignature": {
        const key = tryGetKey(element);
        if (!key)
          continue;
        if (!properties.properties[key] && !properties.methods[key]) {
          const type = (_a = element.typeAnnotation) == null ? void 0 : _a.typeAnnotation;
          properties.properties[key] = {
            value: type ? { type, file } : null,
            optional: !!element.optional,
            signature: { type: element, file }
          };
        }
        break;
      }
      case "TSIndexSignature":
        break;
    }
  }
  return properties;
}
async function resolveTSReferencedType({
  file,
  type
}) {
  var _a;
  if (type.type === "TSTypeAliasDeclaration" || type.type === "TSParenthesizedType") {
    return resolveTSReferencedType({
      file,
      type: type.typeAnnotation
    });
  }
  let refName;
  if (type.type === "Identifier") {
    refName = type.name;
  } else if (type.type === "TSTypeReference") {
    if (type.typeName.type !== "Identifier")
      return void 0;
    refName = type.typeName.name;
  } else {
    return { file, type };
  }
  for (let node of file.ast) {
    if (node.type === "ImportDeclaration") {
      const resolved = await resolveTSFileId(node.source.value, file.filePath);
      if (!resolved)
        continue;
      const specifier = node.specifiers.find(
        (specifier2) => specifier2.type === "ImportSpecifier" && specifier2.imported.type === "Identifier" && specifier2.imported.name === refName
      );
      if (!specifier)
        continue;
      const exports = await resolveTSFileExports(await getTSFile(resolved));
      return exports[specifier.local.name];
    }
    if (node.type === "ExportNamedDeclaration" && node.declaration)
      node = node.declaration;
    if (isTSDeclaration(node)) {
      if (((_a = node.id) == null ? void 0 : _a.type) !== "Identifier")
        continue;
      if (node.id.name !== refName)
        continue;
      return resolveTSReferencedType({
        file,
        type: node
      });
    }
  }
  if (type.type === "TSTypeReference")
    return { file, type };
}
var tsFileExportsCache = /* @__PURE__ */ new Map();
async function resolveTSFileExports(file) {
  var _a;
  if (tsFileExportsCache.has(file))
    return tsFileExportsCache.get(file);
  const exports = {};
  tsFileExportsCache.set(file, exports);
  for (const stmt of file.ast) {
    if (stmt.type === "ExportDefaultDeclaration") {
    } else if (stmt.type === "ExportAllDeclaration") {
      const resolved = await resolveTSFileId(stmt.source.value, file.filePath);
      if (!resolved)
        continue;
      const sourceExports = await resolveTSFileExports(
        await getTSFile(resolved)
      );
      Object.assign(exports, sourceExports);
    } else if (stmt.type === "ExportNamedDeclaration") {
      let sourceExports;
      if (stmt.source) {
        const resolved = await resolveTSFileId(stmt.source.value, file.filePath);
        if (!resolved)
          continue;
        sourceExports = await resolveTSFileExports(await getTSFile(resolved));
      }
      for (const specifier of stmt.specifiers) {
        if (specifier.type === "ExportDefaultSpecifier" || specifier.type === "ExportNamespaceSpecifier") {
          continue;
        }
        const exportedName = specifier.exported.type === "Identifier" ? specifier.exported.name : specifier.exported.value;
        if (stmt.source) {
          exports[exportedName] = sourceExports[specifier.local.name];
        } else {
          exports[exportedName] = await resolveTSReferencedType({
            file,
            type: specifier.local
          });
        }
      }
      if (isTSDeclaration(stmt.declaration)) {
        const decl = stmt.declaration;
        if (((_a = decl.id) == null ? void 0 : _a.type) === "Identifier") {
          const exportedName = decl.id.name;
          exports[exportedName] = await resolveTSReferencedType({
            file,
            type: decl
          });
        }
      }
    }
  }
  return exports;
}
var resolveTSFileIdImpl = resolveTSFileIdNode;
function resolveTSFileId(id, importer) {
  return resolveTSFileIdImpl(id, importer);
}
function resolveTSFileIdNode(id, importer) {
  function tryResolve(id2, importer2) {
    const filePath = path.resolve(importer2, "..", id2);
    if (!existsSync(filePath))
      return;
    return filePath;
  }
  return tryResolve(id, importer) || tryResolve(`${id}.ts`, importer) || tryResolve(`${id}/index`, importer) || tryResolve(`${id}/index.ts`, importer);
}
function setResolveTSFileIdImpl(impl) {
  resolveTSFileIdImpl = impl;
}

export {
  tsFileCache,
  getTSFile,
  isTSDeclaration,
  mergeTSProperties,
  resolveTSProperties,
  resolveTypeElements,
  resolveTSReferencedType,
  tsFileExportsCache,
  resolveTSFileExports,
  resolveTSFileId,
  resolveTSFileIdNode,
  setResolveTSFileIdImpl
};
